#include <FastLED.h>
#include <DS3232RTC.h>
#include <Button.h>
#define LED_PIN     6
#define NUM_LEDS    145
#define BRIGHTNESS  5
DS3232RTC myRTC(false);
tmElements_t tm;
CRGB leds[NUM_LEDS];
CHSV colour( 0, 255, 180);

//Button variables
Button cyclebutton(7);

//various bits for nonblocking code
#define framerate 33 //how many milliseconds to wait between updating the display
unsigned long previousdisplaymillis = 0; //This stores the time when the LED display was last updated
#define backgroundframerate 50 // how many ms to wait between updating the background animation
unsigned long previousbackgroundmillis = 0; //When was the background last updated
#define timeupdate 1000 //how many ms to wait before checking & updating the time
unsigned long previoustimemillis = 0;
unsigned long currentmillis = 0;
unsigned long backgroundtime = 0;
#define backgroundduration 2000

//the LED display translaton array, first number is the row number, second the column number; values are the corresponding LED light for each of tehse positions
const byte LEDarray[23][9] =  {
  {145, 145, 145, 145, 0, 145, 145, 145, 145} ,
  {145, 145, 1, 2, 3, 4, 5, 145, 145} ,
  {145, 12, 11, 10, 9, 8, 7, 6, 145},
  {145, 13, 14, 15, 16, 17, 18, 19, 145} ,
  {145, 26, 25, 24, 23, 22, 21, 20, 145} ,
  {145, 27, 28, 29, 30, 31, 32, 33, 145} ,
  {145, 40, 39, 38, 37, 36, 35, 34, 145} ,
  {145, 41, 42, 43, 44, 45, 46, 47, 145} ,
  {145, 54, 53, 52, 51, 50, 49, 48, 145} ,
  {145, 55, 56, 57, 58, 59, 60, 61, 145} ,
  {145, 68, 67, 66, 65, 64, 63, 62, 145} ,
  {145, 69, 70, 71, 72, 73, 74, 75, 145} ,
  {145, 82, 81, 80, 79, 78, 77, 76, 145} ,
  {145, 83, 84, 85, 86, 87, 88, 89, 145} ,
  {145, 96, 95, 94, 93, 92, 91, 90, 145} ,
  {145, 97, 98, 99, 100, 101, 102, 103, 145} ,
  {145, 110, 109, 108, 107, 106, 105, 104, 145} ,
  {145, 111, 112, 113, 114, 115, 116, 117, 145} ,
  {145, 124, 123, 122, 121, 120, 119, 118, 145} ,
  {145, 125, 126, 127, 128, 129, 130, 131, 145} ,
  {145, 138, 137, 136, 135, 134, 133, 132, 145} ,
  {145, 145, 139, 140, 141, 142, 143, 145, 145} ,
  {145, 145, 145, 145, 144, 145, 145, 145, 145}
};

// The font used for the digit display
const int Digits[10][15] = //each digit (0-9) is a list of the relative sprite positions to light up (3x5 sprite)
{
  {6, 2, 12, 3, 13, 4, 14, 10}, //0
  {6, 2, 7, 8, 9, 5, 10, 15}, //1
  {1, 6, 12, 8, 4, 5, 10, 15}, //2
  {1, 6, 0, 12, 8, 0, 14, 5, 10}, //3
  {11, 2, 12, 3, 13, 4, 9, 14, 15}, //4
  {1, 6, 11, 2, 3, 8, 13, 14, 5, 10}, //5
  {6, 2, 3, 8, 4, 14, 10}, //6
  {1, 6, 11, 12, 8, 9, 10}, //7
  {6, 2, 12, 8, 4, 14, 10}, //8
  {6, 2, 12, 8, 13, 14, 10} //9
};

//Variables used to display the time
uint8_t hr10;
uint8_t hr1;
uint8_t min10;
uint8_t min1;

//Variables to display sprites
int LEDposx = 12; //starting x position for the LED
int LEDposy = 4; //starting y posiiton for the LED

//Variables used for various background routines
static uint8_t theta = 0;
float timeval = 0.0000001;
byte backgroundcounter;
#define stripethickness 15
byte backgroundnumber = 0;

void setup()
{
  Serial.begin (9600); //debug
  cyclebutton.begin();
  delay(1000);
  pinMode(17, OUTPUT); //ground and v5 for clock module
  pinMode(16, OUTPUT);
  digitalWrite(17, HIGH);
  digitalWrite(16, LOW);
  FastLED.addLeds<WS2812, LED_PIN, GRB>(leds, NUM_LEDS);
  myRTC.begin();
}

void loop() {
  currentmillis = millis();
  if (cyclebutton.pressed()) {
    Serial.println("cycle background");
    backgroundnumber++;
    if (backgroundnumber >= 5) backgroundnumber = 0;
  }

  if ((currentmillis - previousdisplaymillis) >= framerate) {
    FastLED.show();
    FastLED.clear();
    RTC.read(tm);
    if (backgroundnumber == 1) {
      BWbands();
    }
    if (backgroundnumber == 2) {
      Bands();
    }
    if (backgroundnumber == 3) {
      PacMan();
    }
    if (backgroundnumber == 4) {
      Twirl();
    }
    timeval = millis();
    backgroundcounter++;
    if (backgroundcounter >= stripethickness) backgroundcounter = 0;
    LEDposx = 11.5 + 6 * (sin(timeval / 1000)); //moves the centre around
    LEDposy = 4.5 + 2 * (cos(timeval / 1000));
    theta = theta + 10;
    showtime();
    previousdisplaymillis = millis();
  }
}

void showtime() { //displays the time
  RTC.read(tm);
  hr10 = tm.Hour / 10;
  hr1 = tm.Hour % 10;
  min10 = tm.Minute / 10;
  min1 = tm.Minute % 10;
  displaynumbersprite(3, hr10);
  displaynumbersprite(7, hr1);
  displaynumbersprite(13, min10);
  displaynumbersprite(17, min1);
  previoustimemillis = millis();
}

void displaynumbersprite(uint8_t place, uint8_t number) { //Used to light up the individual LEDs in a digit
  //light up the LEDs for the current digit
  for (uint8_t i = 0; i < 15; i++) {
    uint8_t value = Digits[number][i];
    value--;
    uint8_t x = (value / 5) + place;
    uint8_t y = (value % 5) + 2;
    if (value > 200) { //pixels with a value of 0 end up at 200, which is outside the display
      x = 0;
      y = 0;
    }
    uint8_t r = (LEDarray [x][y]);
    leds[r] = -leds[r];
    //leds[r] = CRGB::White;
  }

  //display colons
  if ( tm.Second % 2 == 0 ) {
    leds[71] = -leds[71];
    leds[73] = -leds[73];
  }
  else {
    leds[71] = CRGB::White;
    leds[73] = CRGB::White;
  }
}

void BWbands() {   // routine was found in https://stackoverflow.com/questions/3365171/calculating-the-angle-between-two-lines-without-having-to-calculate-the-slope
  //Serial.println("BWBands");
  for (byte i = 0; i <= 22; i++) {
    for (byte j = 0; j <= 8; j++) {
      byte stripeval = ((backgroundcounter + i + j) % stripethickness);
      int r = (LEDarray [i][j]);
      leds[r] = CRGB (stripeval, stripeval, stripeval);
    }
  }
}

void Bands() {   // routine was found in https://stackoverflow.com/questions/3365171/calculating-the-angle-between-two-lines-without-having-to-calculate-the-slope
  //Serial.println("Bands");
  // i and ja are the column and row numbers. These are added together, then the modulo with the stripethickness of the result is taken. This results in diagonally incrementing numbers scaled from 0 to the
  // 'stripethickness' value. These are animatied by adding a counter value, then used to colour the LEDs
  for (byte i = 0; i <= 22; i++) {
    for (byte j = 0; j <= 8; j++) {
      byte modulo = (i + j + backgroundcounter) % stripethickness;
      modulo = map(modulo, 0, stripethickness, 0, 2); //reduces the 'bit depth'
      modulo = map(modulo, 0, 1, 0, 255); //scales it back to full
      int r = (LEDarray [i][j]);
      leds[r] = CRGB (modulo, modulo / 3, 255 - modulo); //when modulo = 255 this gives colour 255,127,0 = orange; when it is 0, it is 0,0,255 = blue
    }
  }
}

void PacMan() {   // routine was found in https://stackoverflow.com/questions/3365171/calculating-the-angle-between-two-lines-without-having-to-calculate-the-slope
  //Serial.println("PacMan");
  for (byte i = 0; i <= 22; i++) {
    for (byte j = 0; j <= LEDposy; j++) {
      int r = (LEDarray [i][j]);
      int dx1 = i - LEDposx;
      int dy1 = j - LEDposy;
      float l2 = (dx1 * dx1 + dy1 * dy1);
      float angle2 = acos (dx1 / sqrt(l2)); //gives the angle of pixel (i,j) with the horizontal (1,0) in radians
      int angle3 = angle2 * 40.585; // scale the radian value to integer 0-255 (where 255 corresponds to 2 radians = 3.1415)
      int delta = theta - angle3; //what is the difference of the pixel's angle and the theta value
      if (delta >= 0 && delta < 127) {
        leds[r] = CRGB(255, 0, 255);
      }
      else {
        leds[r] = CRGB(0, 255, 0);
      }
    }
    for (byte j = LEDposy; j <= 8; j++) {
      int r = (LEDarray [i][j]);
      int dx1 = i - LEDposx;
      int dy1 = j - LEDposy;
      float l2 = (dx1 * dx1 + dy1 * dy1);
      float angle2 = acos (dx1 / sqrt(l2)); //gives the angle of pixel (i,j) with the horizontal (1,0) in radians
      int angle3 = angle2 * 40.585; //
      //angle3 = 255 - angle3;
      int delta = theta - angle3; //what is the difference of the pixel's angle and the theta value
      if (delta >= 0 && delta < 127) {
        leds[r] = CRGB(255, 0, 255);
      }
      else {
        leds[r] = CRGB(0, 255, 0);
      }
    }
    // leds[LEDarray[LEDposx][LEDposy]] = CRGB(0, 255, 255); // the centre LED was misbehaving (presumably rounding issues or similar), needs to be manually set on
  }
}

void Twirl() {   // routine was found in https://stackoverflow.com/questions/3365171/calculating-the-angle-between-two-lines-without-having-to-calculate-the-slope
  //Serial.println("Twirl");
  for (byte i = 0; i <= 22; i++) {
    for (byte j = 0; j <= LEDposy; j++) {
      int r = (LEDarray [i][j]);
      int dx1 = i - LEDposx;
      int dy1 = j - LEDposy;
      float l2 = (dx1 * dx1 + dy1 * dy1);
      float angle2 = acos (dx1 / sqrt(l2)); //gives the angle of pixel (i,j) with the horizontal (1,0) in radians
      int angle3 = angle2 * 40.585; // scale the radian value to integer 0-255 (where 255 corresponds to 2 radians = 3.1415)
      int delta = theta - angle3; //what is the difference of the pixel's angle and the theta value
      if (delta >= 0 && delta < 127) {
        leds[r] = CRGB::Aqua;
      }
      else {
        leds[r] = CRGB::Yellow;
      }
    }
    for (byte j = LEDposy; j <= 8; j++) {
      int r = (LEDarray [i][j]);
      int dx1 = i - LEDposx;
      int dy1 = j - LEDposy;
      float l2 = (dx1 * dx1 + dy1 * dy1);
      float angle2 = acos (dx1 / sqrt(l2)); //gives the angle of pixel (i,j) with the horizontal (1,0) in radians
      int angle3 = angle2 * 40.585; //
      angle3 = 256 - angle3;
      int delta = theta - angle3; //what is the difference of the pixel's angle and the theta value
      leds[r] = CRGB::Yellow;
      if (delta >= 0 && delta < 127) {
        leds[r] = CRGB::Aqua;
      }
      if (delta >= -255 && delta < -127) {
        leds[r] = CRGB::Aqua;
      }
    }
  }
}
